#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Valid parameters for cosmic ray masks computation and application."""

VALID_LACOSMIC_CLEANTYPE = ["median", "medmask", "meanmask", "idw"]
VALID_CRMETHODS = [
    "lacosmic",
    "mm_lacosmic",
    "pycosmic",
    "mm_pycosmic",
    "deepcr",
    "mm_deepcr",
    "conn",
    "mm_conn",
]
VALID_BOUNDARY_FITS = ["spline", "piecewise"]
VALID_COMBINATIONS = ["mean", "median", "min", "mediancr", "meancrt", "meancr", "meancr2"]

DEFAULT_WEIGHT_FIXED_POINTS_IN_BOUNDARY = 10000.0



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Apply thresholding to CR mask"""

from scipy import ndimage


def apply_threshold_cr(bool_crmask2d, bool_threshold2d):
    """Apply thresholding to CR mask.

    Apply threshold only when none of the pixels within a
    grouped CR feature exceeds the threshold condition.
    """

    new_bool_crmask2d = bool_crmask2d.copy()
    structure = [[1, 1, 1], [1, 1, 1], [1, 1, 1]]
    cr_labels, num_crs = ndimage.label(bool_crmask2d, structure=structure)
    for i in range(1, num_crs + 1):
        mask_cr = cr_labels == i
        if not bool_threshold2d[mask_cr].any():
            new_bool_crmask2d[mask_cr] = False

    return new_bool_crmask2d



#
# Copyright 2025-2026 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#
"""Compute RSS image from detector image.

See notes in ifu_simulator.py about parallelisation. 
In this function, the same approach is used as in ifu_simulator.py:
"""

import logging
from multiprocessing import Pool
import numpy as np
import time

from numina.instrument.simulation.ifu.define_3d_wcs import get_wvparam_from_wcs3d

from .update_image2d_rss_method1 import update_image2d_rss_method1


_worker_inputs_method1 = {}

def _worker_init_method1(image2d_detector_method0, 
                         dict_ifu2detector,
                         naxis1_detector, 
                         naxis1_ifu, 
                         nslices,
                         wv_crpix1, 
                         wv_crval1, 
                         wv_cdelt1):
    """Initialiser executed once per worker process at Pool startup.

    Stores all large input arrays in the module-level global so that
    individual tasks only need to receive the slice index.
    """
    global _worker_inputs_method1
    _worker_inputs_method1 = {
        "image2d_detector_method0": image2d_detector_method0,
        "dict_ifu2detector": dict_ifu2detector,
        "naxis1_detector": naxis1_detector,
        "naxis1_ifu": naxis1_ifu,
        "nslices": nslices,
        "wv_crpix1": wv_crpix1,
        "wv_crval1": wv_crval1,
        "wv_cdelt1": wv_cdelt1,
    }


def _worker_method1(islice):
    """Worker function for parallel computation of a single slice.
    
    All large input arrays are accessed from the global _worker_inputs_method1. 
    Allocates private output arrays, calls the update function, 
    and returns results."""
    # Call the update function to compute the RSS slice for the considered slice.
    local_rss = update_image2d_rss_method1(
        islice=islice,
        image2d_detector_method0=_worker_inputs_method1['image2d_detector_method0'],
        dict_ifu2detector=_worker_inputs_method1['dict_ifu2detector'],
        naxis1_detector=_worker_inputs_method1['naxis1_detector'],
        naxis1_ifu=_worker_inputs_method1['naxis1_ifu'],
        wv_crpix1=_worker_inputs_method1['wv_crpix1'],
        wv_crval1=_worker_inputs_method1['wv_crval1'],
        wv_cdelt1=_worker_inputs_method1['wv_cdelt1'],
        debug=False,
        logger=None,
    )
    return islice, local_rss


def compute_image2d_rss_from_detector_method1(
    image2d_detector_method0,
    naxis1_detector,
    naxis1_ifu,
    nslices,
    dict_ifu2detector,
    wcs3d,
    ncores=1,
    logger=None,
    console=None,
):
    """
    Compute the RSS image from the detector image using method 1.

    Parameters
    ----------
    image2d_detector_method0 : numpy.ndarray
        The input detector image.
    naxis1_detector : `~astropy.units.Quantity`
        Number of pixels in the first axis of the detector image
        (dispersion direction)
    naxis1_ifu : `~astropy.units.Quantity`
        Number of pixels in the first axis of the IFU image.
    nslices : int
        Number of slices in the IFU image.
    dict_ifu2detector : dict
        Dictionary mapping IFU slices to detector pixels.
    wcs3d : `~astropy.wcs.WCS`
        WCS object for the 3D image.
    ncores : int, optional
        Number of CPU cores to be used for parallel processing. If ncores=1,
        no parallel processing is used. If ncores > 1, the computation is
        parallelised using multiprocessing.Pool. Default is 1 (no parallelisation).
    verbose : bool, optional
        If True, print verbose output. Default is False.
    logger : `~logging.Logger`, optional
        Logger for logging messages. If None, the root logger is used.
    console : `~rich.console.Console` or None, optional
        Rich console for rich printing. If None, the default console will be used.

    Returns
    -------
    image2d_rss_method1 : numpy.ndarray
        The computed RSS image.
    """

    if logger is None:
        logger = logging.getLogger()
    logger_level_in_use = logger.getEffectiveLevel()

    logger.debug("[green]* Computing image2d RSS (method 1)[/green]")

    # Get WCS parameters
    wv_cunit1, wv_crpix1, wv_crval1, wv_cdelt1 = get_wvparam_from_wcs3d(wcs3d)

    # Initialize image
    image2d_rss_method1 = np.zeros((naxis1_ifu.value * nslices, naxis1_detector.value))

    logger.debug("Rectifying...")
    t0 = time.time()

    if ncores == 1:
        # Explicit loop in slices
        for islice in range(nslices):
            if logger_level_in_use <= logging.DEBUG:
                logger.debug(f"slice: {islice + 1}/{nslices}")
            else:
                if (islice + 1) % 10 == 0:
                    console.print(f"{islice + 1}", end="")
                else:
                    console.print(".", end="")
            single_slice2d = update_image2d_rss_method1(
                islice=islice,
                image2d_detector_method0=image2d_detector_method0,
                dict_ifu2detector=dict_ifu2detector,
                naxis1_detector=naxis1_detector,
                naxis1_ifu=naxis1_ifu,
                wv_crpix1=wv_crpix1,
                wv_crval1=wv_crval1,
                wv_cdelt1=wv_cdelt1,
                debug=False,
                logger=logger,
            )
            # insert slice in final image
            i1_rss_index = islice * naxis1_ifu.value
            i2_rss_index = i1_rss_index + naxis1_ifu.value
            image2d_rss_method1[i1_rss_index:i2_rss_index, :] += single_slice2d[:, :]
        if logger_level_in_use > logging.DEBUG:
            console.print("")  # new line after the progress dots
    else:
        # Large input arrays are loaded into each worker once via the initializer,
        # so only the slice index is passed per task, minimising serialisation overhead.
        with Pool(
            processes=ncores,
            initializer=_worker_init_method1,
            initargs=(
                image2d_detector_method0,
                dict_ifu2detector,
                naxis1_detector,
                naxis1_ifu,
                nslices,
                wv_crpix1,
                wv_crval1,
                wv_cdelt1,
            ),
        ) as pool:
            # insert slices in final image as they are completed
            for islice, local_rss in pool.imap_unordered(_worker_method1, range(nslices)):
                i1_rss_index = islice * naxis1_ifu.value
                i2_rss_index = i1_rss_index + naxis1_ifu.value
                image2d_rss_method1[i1_rss_index:i2_rss_index, :] = local_rss[:, :]

    t1 = time.time()
    logger.info(f"Delta time: {t1 - t0}")

    return image2d_rss_method1



#
# Copyright 2025-2026 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Execute PyCosmic cosmic ray detection algorithm."""

from datetime import datetime
import numpy as np

try:
    import deepCR

    DEEPCR_AVAILABLE = True
except ModuleNotFoundError as e:
    DEEPCR_AVAILABLE = False


def execute_deepcr(image2d, bool_to_be_cleaned, rlabel_deepcr, dict_dc_params, displaypar, _logger):
    """Execute DeepCR cosmic ray detection algorithm."""
    if not DEEPCR_AVAILABLE:
        raise ImportError(
            "DeepCR is not installed. Please install DeepCR to use\n"
            "the 'deepcr' or 'mm_deepcr' crmethod options.\n"
            "You can try installing it via pip:\n"
            "pip install deepCR\n"
        )
    if displaypar:
        _logger.info("[green][DeepCR parameters][/green]")
        for key in dict_dc_params.keys():
            _logger.info("%s for deepCR: %s", key, str(dict_dc_params[key]))
    # Execute DeepCR
    _logger.info("using deepCR version: %s", deepCR.__version__)
    time_ini = datetime.now()
    mdl = deepCR.deepCR(mask=dict_dc_params["mask"])
    flag_dc, median2d_deepcr = mdl.clean(  # note the order of outputs!
        image2d,
        threshold=dict_dc_params["threshold"],
        inpaint=True,
    )
    flag_dc = np.logical_and(flag_dc, bool_to_be_cleaned)
    flag_dc = flag_dc.flatten()
    datetime_end = datetime.now()
    delta_datetime = datetime_end - time_ini
    _logger.info("deepCR execution time: %s", str(delta_datetime))
    _logger.info(
        "pixels flagged as cosmic rays by %s: %d (%08.4f%%)",
        rlabel_deepcr,
        np.sum(flag_dc),
        np.sum(flag_dc) / flag_dc.size * 100,
    )

    return median2d_deepcr, flag_dc



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#
"""Decorator to capture stdout and stderr of a function and log it."""

import io
import logging

from contextlib import redirect_stderr, redirect_stdout
from functools import wraps


def decorate_output(prompt=">>>"):
    """Decorator factory to customize the prompt."""

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            _logger = logging.getLogger(__name__)
            buf = io.StringIO()
            with redirect_stdout(buf), redirect_stderr(buf):
                result = func(*args, **kwargs)
            # Split into lines
            output_lines = buf.getvalue().splitlines()
            # Remove trailing empty line
            if output_lines and output_lines[-1] == "":
                output_lines = output_lines[:-1]
            if output_lines:
                prefixed_lines = [f"{prompt}{line}" for line in output_lines]
                _logger.info("\n".join(prefixed_lines))
            return result

        return wrapper

    return decorator



#
# Copyright 2024-2026 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#
"""IFU simulator

We have avoided using joblib.Parallel for the computation because
when using "prefer=threads" in joblib, multiple threads write 
simultaneously to the same shared NumPy arrays 
(image2d_rss_method0, image2d_detector_method0). 
Python's GIL does not protect NumPy array writes, so concurrent modifications 
produce corrupted pixel values that change randomly between runs — the classic 
signature of a race condition.

The solution: private arrays + final reduction:
The pattern applied in both worker functions (here and in 
compute_image2d_rss_from_detector_method1) is identical:
1. Each worker allocates its own private output arrays (local_rss, local_det) 
   initialised to zero, works on them independently, and returns them.
2. Pool.map distributes the slice tasks across physical CPU cores using 
   separate processes. Since each process has its own memory space, 
   there is no shared state and no race condition, regardless of whether 
   slices write to overlapping regions.
3. The main process reduces the results by accumulating all returned 
   local arrays into the final output arrays with +=. 
   This step is single-threaded and therefore always safe.

The remaining performance bottleneck is the serialisation overhead 
of passing large NumPy arrays to each worker on every call. 
The proposed optimisation is to use the initializer argument of Pool 
to load the large read-only input arrays into each worker process once 
at startup, so that only the slice index needs to be passed per task. 
This reduces both serialisation time and memory pressure, and is 
the most promising avenue for closing the gap between parallel and 
sequential execution times.

The timing results show that this workload is memory-bound rather 
than compute-bound. Each slice computation is fast enough that 
the bottleneck is not the CPU but the rate at which data can be 
fetched from RAM. As a consequence, adding more parallel processes 
beyond a small number does not help — it simply increases contention
on the memory bus, and the overhead of managing additional processes
begins to dominate. In fact, in some cases using more than one
process can even degrade performance due to this increased contention
and overhead.
"""

from astropy.io import fits
from astropy.units import Quantity
import astropy.units as u
import json
import logging
from multiprocessing import Pool
import numpy as np
import os
from pathlib import Path
from rich.console import Console
from rich.pretty import pretty_repr
import time
import yaml

from .define_3d_wcs import get_wvparam_from_wcs3d
from .define_3d_wcs import wcs_to_header_using_cd_keywords
from .display_skycalc import display_skycalc
from .compute_image2d_rss_from_detector_method1 import compute_image2d_rss_from_detector_method1
from .compute_image3d_ifu_from_rss_method1 import compute_image3d_ifu_from_rss_method1
from .generate_image2d_method0_ifu import generate_image2d_method0_ifu
from .generate_geometry_for_scene_block import generate_geometry_for_scene_block
from .generate_image3d_method0_ifu import generate_image3d_method0_ifu
from .generate_spectrum_for_scene_block import generate_spectrum_for_scene_block
from .load_atmosphere_transmission_curve import load_atmosphere_transmission_curve
from .raise_valueerror import raise_ValueError
from .save_image2d_detector_method0 import save_image2d_detector_method0
from .save_image2d_rss import save_image2d_rss
from .set_wavelength_unit_and_range import set_wavelength_unit_and_range
from .update_image2d_rss_detector_method0 import update_image2d_rss_detector_method0


_worker_inputs_method0 = {}

def _worker_init_method0(simulated_x_ifu_all,
                         simulated_y_ifu_all,
                         simulated_wave_all,
                         naxis1_ifu,
                         naxis1_detector,
                         naxis2_detector,
                         nslices,
                         bins_x_ifu,
                         bins_wave,
                         bins_x_detector,
                         bins_y_detector,
                         wv_cdelt1,
                         extra_degradation_spectral_direction,
                         dict_ifu2detector):
    """Initialiser executed once per worker process at Pool startup.
    
    Stores all large input arrays in the module-level global so that
    individual tasks only need to receive the slice index.
    """
    global _worker_inputs_method0
    _worker_inputs_method0 = {
        "simulated_x_ifu_all": simulated_x_ifu_all,
        "simulated_y_ifu_all": simulated_y_ifu_all,
        "simulated_wave_all": simulated_wave_all,
        "naxis1_ifu": naxis1_ifu,
        "naxis1_detector": naxis1_detector,
        "naxis2_detector": naxis2_detector,
        "nslices": nslices,
        "bins_x_ifu": bins_x_ifu,
        "bins_wave": bins_wave,
        "bins_x_detector": bins_x_detector,
        "bins_y_detector": bins_y_detector,
        "wv_cdelt1": wv_cdelt1,
        "extra_degradation_spectral_direction": extra_degradation_spectral_direction,
        "dict_ifu2detector": dict_ifu2detector
    }

def _worker_method0(islice):
    """Worker function for parallel computation of a single slice.
    
    Receives only the slice index, and accesses all large input arrays 
    from the global _worker_inputs_method0. Allocates private output arrays, 
    calls the update function, and returns results."""
    local_rss = np.zeros(
        (_worker_inputs_method0['naxis1_ifu'].value * _worker_inputs_method0['nslices'],
        _worker_inputs_method0['naxis1_detector'].value), 
        dtype=int
    )
    local_detector = np.zeros(
        (_worker_inputs_method0['naxis2_detector'].value, 
         _worker_inputs_method0['naxis1_detector'].value))  # float type to be able to include noise and flatfield effects

    update_image2d_rss_detector_method0(
        islice=islice,
        simulated_x_ifu_all=_worker_inputs_method0['simulated_x_ifu_all'],
        simulated_y_ifu_all=_worker_inputs_method0['simulated_y_ifu_all'],
        simulated_wave_all=_worker_inputs_method0['simulated_wave_all'],
        naxis1_ifu=_worker_inputs_method0['naxis1_ifu'],
        bins_x_ifu=_worker_inputs_method0['bins_x_ifu'],
        bins_wave=_worker_inputs_method0['bins_wave'],
        bins_x_detector=_worker_inputs_method0['bins_x_detector'],
        bins_y_detector=_worker_inputs_method0['bins_y_detector'],
        wv_cdelt1=_worker_inputs_method0['wv_cdelt1'],
        extra_degradation_spectral_direction=_worker_inputs_method0['extra_degradation_spectral_direction'],
        dict_ifu2detector=_worker_inputs_method0['dict_ifu2detector'],
        image2d_rss_method0=local_rss,
        image2d_detector_method0=local_detector,
    )
    return local_rss, local_detector


def ifu_simulator(
    wcs3d,
    header_keys,
    naxis1_detector,
    naxis2_detector,
    nslices,
    noversampling_whitelight,
    scene_fname,
    flux_factor,
    seeing_fwhm_arcsec,
    seeing_psf,
    instrument_pa,
    airmass,
    parallactic_angle,
    flatpix2pix,
    atmosphere_transmission,
    bias,
    rnoise,
    spectral_blurring_pixel,
    bitpix_detector,
    faux_dict,
    rng,
    ncores=1,
    prefix_intermediate_fits='test',
    stop_after_ifu_3D_method0=False,
    logger=None,
    console=None,
    instname=None,
    subtitle=None,
    plots=False,
    output_dir=".",
):
    """IFU simulator.

    This function does not return any value. The main output of this function
    is a set of FITS files with intermediate results, which are stored in
    the output directory defined by the user.

    Parameters
    ----------
    wcs3d : `~astropy.wcs.wcs.WCS`
        WCS of the data cube.
    header_keys : `~astropy.io.fits.header.Header`
        FITS header with additional keywords to be merged together with
        the WCS information.
    naxis1_detector : `~astropy.units.Quantity`
        Detector NAXIS1, dispersion direction.
    naxis2_detector : `~astropy.units.Quantity`
        Detector NAXIS2, spatial direction (slices).
    nslices : int
        Number of IFU slices.
    noversampling_whitelight : int
        Oversampling factor (integer number) to generate the white
        light image.
    scene_fname : str
        YAML scene file name.
    flux_factor : float
        Flux factor to be applied to the number of photons defined
        in the scene file.
    seeing_fwhm_arcsec : `~astropy.units.Quantity`
        Seeing FWHM.
    seeing_psf : str
        Seeing PSF.
    instrument_pa : `~astropy.units.Quantity`
        Instrument position angle.
    airmass : float
        Airmass.
    parallactic_angle : `~astropy.units.Quantity`
        Parallactic angle. This number must be within the range
        [-90,+90] deg.
    flatpix2pix : str
        String indicating whether a pixel-to-pixel flatfield is
        applied or not. Two possible values:
        - 'default': use default flatfield defined in 'faux_dict'
        - 'none': do not apply flatfield
    atmosphere_transmission : str
        String indicating whether the atmosphere transmission of
        the atmosphere is applied or not. Two possible values are:
        - 'default': use default curve defined in 'faux_dict'
        - 'none': do not apply atmosphere transmission
    bias : `~astropy.units.Quantity`
        Bias level (in ADU).
    rnoise : `~astropy.units.Quantity`
        Readout noise standard deviation (in ADU). Assumed to be
        Gaussian.
    spectral_blurring_pixel : `~astropy.units.Quantity`
        Spectral blurring (in pixels) to be introduced when generating
        the initial RSS image from the original 3D data cube.
    bitpix_detector : int
        BITPIX value for the detector image. Supported values are:
        - 16: unsigned short integer (0 to 65535)
        - -32: 32-bit floating point
    faux_dict : Python dictionary
        File names of auxiliary files:
        - skycalc: table with SKYCALC Sky Model Calculator predictions
        - flatpix2pix: pixel-to-pixel flat field
        - model_ifu2detector: 2D polynomial transformation
          x_ifu, y_ify, wavelength -> x_detector, y_detector
    rng : `~numpy.random._generator.Generator`
        Random number generator.
    ncores : int
        Number of CPU cores to be used for parallel processing. If ncores=1,
        no parallel processing is used. If ncores > 1, the computation is
        parallelised using multiprocessing.Pool. Default is 1 (no parallelisation).
    prefix_intermediate_fits : str
        Prefix for output intermediate FITS files. If the length of
        this string is 0, no output is generated.
    stop_after_ifu_3D_method0 : bool
        If True, stop execution after generating ifu_3D_method0 image.
    logger : logging.Logger or None, optional
        Logger for logging messages. If None, a default logger will be used.
    console : `~rich.console.Console` or None, optional
        Rich console for rich printing. If None, the default console will be used.
    instname : str or None
        Instrument name.
    subtitle : str or None
        Plot subtitle.
    plots : bool
        If True, plot intermediate results.
    output_dir : str or `~pathlib.Path`, optional
        Output directory to store results.
    """

    if logger is None:
        logger = logging.getLogger(__name__)
    logger_level_in_use = logger.getEffectiveLevel()

    if console is None:
        console = Console()

    if logger_level_in_use <= logging.DEBUG:
        logger.debug(" ")
        for item in faux_dict:
            logger.debug(f"- Required file for item {item}: {faux_dict[item]}")

    # Check bitpix_detector value
    if bitpix_detector not in [16, -32]:
        raise ValueError(f"Unsupported BITPIX value: {bitpix_detector}. Supported values are 16 and -32.")

    # Check output_dir is a valid directory
    if not os.path.isdir(output_dir):
        raise_ValueError(f"Invalid output directory: {output_dir}")
    else:
        logger.debug(f"Output directory: {output_dir}")

    # Check ncores is a positive integer
    if not isinstance(ncores, int) or ncores < 1:
        raise_ValueError(f"Invalid ncores value: {ncores}. ncores must be a positive integer.")
    # If ncores > 1, we will use parallel processing. 
    # We will check the number of available CPU cores and adjust ncores 
    # if it exceeds the available cores or if os.cpu_count() returns None.
    if ncores > 1:
        cpu_count = os.cpu_count()
        if cpu_count is None:
            if ncores > 1:
                logger.warning(f"[cyan]WARNING: os.cpu_count() returned None, using ncores=1 instead[/cyan]")
                ncores = 1
        else:
            if ncores > cpu_count:
                logger.warning(f"[cyan]WARNING: ncores={ncores} exceeds available CPU cores ({cpu_count})[/cyan]")
                ncores = cpu_count - 1
            logger.info(f"os.cpu_count() returned {cpu_count}, using ncores={ncores} for parallel computation")

    if plots:
        # Display SKYCALC predictions for sky radiance and transmission
        display_skycalc(faux_skycalc=faux_dict["skycalc"])

    # Spatial IFU limits
    naxis1_ifu = Quantity(value=wcs3d.array_shape[2], unit=u.pix, dtype=int)
    naxis2_ifu = Quantity(value=wcs3d.array_shape[1], unit=u.pix, dtype=int)
    min_x_ifu = 0.5 * u.pix
    max_x_ifu = naxis1_ifu + 0.5 * u.pix
    min_y_ifu = 0.5 * u.pix
    max_y_ifu = naxis2_ifu + 0.5 * u.pix

    # Wavelength limits
    wv_cunit1, wv_crpix1, wv_crval1, wv_cdelt1 = get_wvparam_from_wcs3d(wcs3d)
    wmin = wv_crval1 + (0.5 * u.pix - wv_crpix1) * wv_cdelt1
    wmax = wmin + naxis1_detector * wv_cdelt1
    reference_wave_vacuum_differential_refraction = (wmin + wmax) / 2

    # Load atmosphere transmission curve
    wave_transmission, curve_transmission = load_atmosphere_transmission_curve(
        atmosphere_transmission=atmosphere_transmission,
        wmin=wmin,
        wmax=wmax,
        wv_cunit1=wv_cunit1,
        faux_dict=faux_dict,
        logger=logger,
    )

    required_keys_in_scene_block = {
        "scene_block_name",
        "spectrum",
        "geometry",
        "nphotons",
        "wavelength_sampling",  # default: random
        "apply_atmosphere_transmission",  # default: True
        "apply_seeing",  # default: True
        "render",  # default: True
    }

    nphotons_all = 0
    simulated_wave_all = None
    simulated_x_ifu_all = None
    simulated_y_ifu_all = None

    # Main loop (rendering of scene blocks)
    with open(scene_fname, "rt") as fstream:
        scene_dict = yaml.safe_load_all(fstream)
        for scene_block in scene_dict:
            if "scene_block_name" not in scene_block.keys():
                logger.info(" ")
                logger.info(f"[red]ERROR while processing {scene_fname}[/red]")
                raise_ValueError("key scene_block_name not found!")
            scene_block_name = scene_block["scene_block_name"]
            logger.info(f"[green]* Processing: {scene_block_name}[/green]")
            # insert default values for keys not provided
            if "wavelength_sampling" not in scene_block.keys():
                scene_block["wavelength_sampling"] = "random"
                logger.info(f"[cyan]WARNING: asumming {scene_block['wavelength_sampling']=}[/cyan]")
            if "render" not in scene_block.keys():
                scene_block["render"] = True
                logger.info(f"[cyan]WARNING: asumming {scene_block['render']=}[/cyan]")
            if "apply_atmosphere_transmission" not in scene_block.keys():
                scene_block["apply_atmosphere_transmission"] = True
                logger.info(f"[cyan]WARNING: asumming {scene_block['apply_atmosphere_transmission']=}[/cyan]")
            if "apply_seeing" not in scene_block.keys():
                scene_block["apply_seeing"] = True
                logger.info(f"[cyan]WARNING: asumming {scene_block['apply_seeing']=}[/cyan]")
            scene_block_keys = set(scene_block.keys())
            # insert default values
            if scene_block_keys != required_keys_in_scene_block:
                logger.info(f"[red]ERROR while processing: {scene_fname}[/red]")
                logger.info(f"[blue]expected keys..: [/blue]{required_keys_in_scene_block}")
                logger.info(f"[blue]keys found.....: [/blue]{scene_block_keys}")
                list_unexpected_keys = list(scene_block_keys.difference(required_keys_in_scene_block))
                if len(list_unexpected_keys) > 0:
                    logger.info(f"[red]unexpected keys: [/red]{list_unexpected_keys}")
                list_missing_keys = list(required_keys_in_scene_block.difference(scene_block_keys))
                if len(list_missing_keys) > 0:
                    logger.info(f"[red]missing keys...: [/red]{list_missing_keys}")
                logger.info(f"scene_block={pretty_repr(scene_block)}")
                raise_ValueError(f"Invalid format in file: {scene_fname}")
            logger.debug(f"scene_block={pretty_repr(scene_block)}")

            nphotons = int(float(scene_block["nphotons"]) * flux_factor)
            logger.info(f"Number of photons to be simulated for this block: {nphotons}")
            wavelength_sampling = scene_block["wavelength_sampling"]
            if wavelength_sampling not in ["random", "fixed"]:
                raise_ValueError(f"Unexpected {wavelength_sampling=}")
            apply_atmosphere_transmission = scene_block["apply_atmosphere_transmission"]
            if atmosphere_transmission == "none" and apply_atmosphere_transmission:
                logger.info(f"[cyan]WARNING: {apply_atmosphere_transmission=} when {atmosphere_transmission=}[/cyan]")
                logger.info(f"{atmosphere_transmission=} overrides {apply_atmosphere_transmission=}")
                logger.info("The atmosphere transmission will not be applied!")
                apply_atmosphere_transmission = False
            apply_seeing = scene_block["apply_seeing"]
            if apply_seeing:
                if seeing_fwhm_arcsec.value < 0:
                    raise_ValueError(f"Unexpected {seeing_fwhm_arcsec=}")
                elif seeing_fwhm_arcsec == 0:
                    logger.info(f"[cyan]WARNING: {apply_seeing=} when {seeing_fwhm_arcsec=}[/cyan]")
                    logger.info("Seeing effect will not be applied!")
                    apply_seeing = False
            render = scene_block["render"]
            if nphotons > 0 and render:
                # set wavelength unit and range
                wave_unit, wave_min, wave_max = set_wavelength_unit_and_range(
                    scene_fname=scene_fname,
                    scene_block=scene_block,
                    wmin=wmin,
                    wmax=wmax,
                    logger=logger,
                )
                # generate spectrum
                simulated_wave = generate_spectrum_for_scene_block(
                    scene_fname=scene_fname,
                    scene_block=scene_block,
                    faux_dict=faux_dict,
                    wave_unit=wave_unit,
                    wave_min=wave_min,
                    wave_max=wave_max,
                    nphotons=nphotons,
                    wavelength_sampling=wavelength_sampling,
                    apply_atmosphere_transmission=apply_atmosphere_transmission,
                    wave_transmission=wave_transmission,
                    curve_transmission=curve_transmission,
                    rng=rng,
                    naxis1_detector=naxis1_detector,
                    logger=logger,
                    plots=plots,
                )
                # convert to default wavelength_unit
                simulated_wave = simulated_wave.to(wv_cunit1)
                # distribute photons in the IFU focal plane
                simulated_x_ifu, simulated_y_ifu = generate_geometry_for_scene_block(
                    scene_fname=scene_fname,
                    scene_block=scene_block,
                    nphotons=nphotons,
                    apply_seeing=apply_seeing,
                    seeing_fwhm_arcsec=seeing_fwhm_arcsec,
                    seeing_psf=seeing_psf,
                    airmass=airmass,
                    parallactic_angle=parallactic_angle,
                    reference_wave_vacuum_differential_refraction=reference_wave_vacuum_differential_refraction,
                    simulated_wave=simulated_wave,
                    instrument_pa=instrument_pa,
                    wcs3d=wcs3d,
                    min_x_ifu=min_x_ifu,
                    max_x_ifu=max_x_ifu,
                    min_y_ifu=min_y_ifu,
                    max_y_ifu=max_y_ifu,
                    nslices=nslices,
                    rng=rng,
                    logger=logger,
                    plots=plots,
                )
                # store all simulated photons
                if nphotons_all == 0:
                    simulated_wave_all = simulated_wave
                    simulated_x_ifu_all = simulated_x_ifu
                    simulated_y_ifu_all = simulated_y_ifu
                else:
                    simulated_wave_all = np.concatenate((simulated_wave_all, simulated_wave))
                    simulated_x_ifu_all = np.concatenate((simulated_x_ifu_all, simulated_x_ifu))
                    simulated_y_ifu_all = np.concatenate((simulated_y_ifu_all, simulated_y_ifu))
                # ---
                # update nphotons
                logger.debug(f"[blue]--> {nphotons} photons simulated[/blue]")
                if nphotons_all == 0:
                    nphotons_all = nphotons
                else:
                    nphotons_all += nphotons
                if (
                    len({nphotons_all, len(simulated_wave_all), len(simulated_x_ifu_all), len(simulated_y_ifu_all)})
                    != 1
                ):
                    logger.info(f"[red]ERROR: check the following numbers:[/red]")
                    logger.info(f"{nphotons_all=}")
                    logger.info(f"{len(simulated_wave_all)=}")
                    logger.info(f"{len(simulated_x_ifu_all)=}")
                    logger.info(f"{len(simulated_y_ifu_all)=}")
                    raise_ValueError("Unexpected differences found in the previous numbers")
            else:
                if nphotons == 0:
                    logger.warning(f"[cyan]WARNING -> nphotons: 0[/cyan]")
                else:
                    logger.warning(f"[cyan]WARNING -> render: False[/cyan]")

    # Filter simulated photons to keep only those that fall within
    # the IFU field of view and within the expected spectral range
    # (note that this step also removes simulated photons with
    # negative wavelength value corresponding to those absorbed by
    # the atmosphere when applying the transmission curve)
    textwidth_nphotons_number = len(str(nphotons_all))
    if nphotons_all > 0:
        logger.debug("Filtering photons within IFU field of view and spectral range...")
        logger.debug(f"Initial number of simulated photons: {nphotons_all:>{textwidth_nphotons_number}}")
        cond1 = simulated_x_ifu_all >= min_x_ifu
        cond2 = simulated_x_ifu_all <= max_x_ifu
        cond3 = simulated_y_ifu_all >= min_y_ifu
        cond4 = simulated_y_ifu_all <= max_y_ifu
        cond5 = simulated_wave_all >= wmin
        cond6 = simulated_wave_all <= wmax
        iok = np.where(cond1 & cond2 & cond3 & cond4 & cond5 & cond6)[0]
    else:
        iok = np.array([], dtype=int)

    if len(iok) < nphotons_all:
        simulated_x_ifu_all = simulated_x_ifu_all[iok]
        simulated_y_ifu_all = simulated_y_ifu_all[iok]
        simulated_wave_all = simulated_wave_all[iok]
        nphotons_all = len(iok)
    logger.debug(f"[blue]Final number of simulated photons..: {nphotons_all:>{textwidth_nphotons_number}}[/blue]")

    # ---------------------------------------------------------------
    # Compute image2d IFU, white image, with and without oversampling
    # ---------------------------------------------------------------
    logger.debug(f"[green]* Computing image2d IFU (method 0) with and without oversampling[/green]")
    for noversampling in {noversampling_whitelight, 1}:
        generate_image2d_method0_ifu(
            wcs3d=wcs3d,
            header_keys=header_keys,
            noversampling_whitelight=noversampling,
            simulated_x_ifu_all=simulated_x_ifu_all,
            simulated_y_ifu_all=simulated_y_ifu_all,
            prefix_intermediate_fits=prefix_intermediate_fits,
            instname=instname,
            subtitle=subtitle,
            scene=scene_fname,
            plots=plots,
            logger=logger,
            output_dir=output_dir,
        )

    # ----------------------------
    # Compute image3d IFU, method0
    # ----------------------------
    logger.debug(f"[green]* Computing image3d IFU (method 0)[/green]")
    bins_x_ifu = (0.5 + np.arange(naxis1_ifu.value + 1)) * u.pix
    bins_y_ifu = (0.5 + np.arange(naxis2_ifu.value + 1)) * u.pix
    bins_wave = (
        wv_crval1
        + ((np.arange(naxis2_detector.value + 1) + 1) * u.pix - wv_crpix1) * wv_cdelt1
        - 0.5 * u.pix * wv_cdelt1
    )
    generate_image3d_method0_ifu(
        wcs3d=wcs3d,
        header_keys=header_keys,
        simulated_x_ifu_all=simulated_x_ifu_all,
        simulated_y_ifu_all=simulated_y_ifu_all,
        simulated_wave_all=simulated_wave_all,
        bins_x_ifu=bins_x_ifu,
        bins_y_ifu=bins_y_ifu,
        bins_wave=bins_wave,
        prefix_intermediate_fits=prefix_intermediate_fits,
        logger=logger,
        output_dir=output_dir,
    )

    if stop_after_ifu_3D_method0:
        raise SystemExit(f"Program stopped because {stop_after_ifu_3D_method0=}")

    # --------------------------------------------
    # Compute image2d RSS and in detector, method0
    # --------------------------------------------
    logger.debug(f"[green]* Computing image2d RSS and detector (method 0)[/green]")
    bins_x_detector = np.linspace(start=0.5, stop=naxis1_detector.value + 0.5, num=naxis1_detector.value + 1)
    bins_y_detector = np.linspace(start=0.5, stop=naxis2_detector.value + 0.5, num=naxis2_detector.value + 1)

    # Read ifu2detector transformations
    dict_ifu2detector = json.loads(open(faux_dict["model_ifu2detector"], mode="rt").read())

    # Additional degradation in the spectral direction
    # (in units of detector pixels)
    if spectral_blurring_pixel.unit != u.pix:
        raise_ValueError(f"Unexpected unit for {spectral_blurring_pixel.unit=}")
    extra_degradation_spectral_direction = (
        rng.normal(loc=0.0, scale=spectral_blurring_pixel.value, size=nphotons_all) * u.pix
    )

    # Initialize images
    image2d_rss_method0 = np.zeros((naxis1_ifu.value * nslices, naxis1_detector.value), dtype=int)
    image2d_detector_method0 = np.zeros(
        (naxis2_detector.value, naxis1_detector.value)
    )  # float type to be able to include noise and flatfield effects

    # Update images
    t0 = time.time()
    if ncores == 1:
        # explicit loop in slices
        for islice in range(nslices):
            if logger_level_in_use <= logging.DEBUG:
                logger.debug(f"slice: {islice + 1}/{nslices}")
            else:
                if (islice + 1) % 10 == 0:
                    console.print(f"{islice + 1}", end="")
                else:
                    console.print(".", end="")
            update_image2d_rss_detector_method0(
                islice=islice,
                simulated_x_ifu_all=simulated_x_ifu_all,
                simulated_y_ifu_all=simulated_y_ifu_all,
                simulated_wave_all=simulated_wave_all,
                naxis1_ifu=naxis1_ifu,
                bins_x_ifu=bins_x_ifu,
                bins_wave=bins_wave,
                bins_x_detector=bins_x_detector,
                bins_y_detector=bins_y_detector,
                wv_cdelt1=wv_cdelt1,
                extra_degradation_spectral_direction=extra_degradation_spectral_direction,
                dict_ifu2detector=dict_ifu2detector,
                image2d_rss_method0=image2d_rss_method0,
                image2d_detector_method0=image2d_detector_method0,
            )
        if logger_level_in_use > logging.DEBUG:
            console.print("")  # new line after the progress dots
    else:
        # Parallel path using multiprocessing.Pool.
        # Large input arrays are loaded into each worker once via the initializer,
        # so only the slice index is passed per task, minimising serialisation overhead.
        # Number of workers is capped at physical core count and at nslices to avoid
        # spawning idle processes.
        cpu_count = os.cpu_count()
        with Pool(
            processes=ncores,
            initializer=_worker_init_method0,
            initargs=(
                simulated_x_ifu_all,
                simulated_y_ifu_all,
                simulated_wave_all,
                naxis1_ifu,
                naxis1_detector,
                naxis2_detector,
                nslices,
                bins_x_ifu,
                bins_wave,
                bins_x_detector,
                bins_y_detector,
                wv_cdelt1,
                extra_degradation_spectral_direction,
                dict_ifu2detector,
            )
        ) as pool:
            results = pool.map(_worker_method0, range(nslices))
        # Final reduction (single-threaded, always safe)
        for local_rss, local_detector in results:
            image2d_rss_method0 += local_rss
            image2d_detector_method0 += local_detector

    t1 = time.time()
    logger.info(f"Delta time: {t1 - t0}")

    # Save RSS image (note that the flatfield effect is not included!)
    save_image2d_rss(
        wcs3d=wcs3d,
        header_keys=header_keys,
        image2d_rss=image2d_rss_method0,
        method=0,
        prefix_intermediate_fits=prefix_intermediate_fits,
        bitpix=16,
        logger=logger,
        output_dir=output_dir,
    )

    # Apply flatpix2pix to detector image
    if flatpix2pix not in ["default", "none"]:
        raise_ValueError(f"Invalid {flatpix2pix=}")
    if flatpix2pix == "default":
        infile = faux_dict["flatpix2pix"]
        with fits.open(infile) as hdul:
            image2d_flatpix2pix = hdul[0].data
        if np.min(image2d_flatpix2pix) <= 0:
            logger.info(f"- minimum flatpix2pix value: {np.min(image2d_flatpix2pix)}")
            raise_ValueError("Unexpected signal in flatpix2pix <= 0")
        naxis2_flatpix2pix, naxis1_flatpix2pix = image2d_flatpix2pix.shape
        naxis1_flatpix2pix *= u.pix
        naxis2_flatpix2pix *= u.pix
        if (naxis1_flatpix2pix != naxis1_detector) or (naxis2_flatpix2pix != naxis2_detector):
            raise_ValueError(f"Unexpected flatpix2pix shape: naxis1={naxis1_flatpix2pix}, naxis2={naxis2_flatpix2pix}")
        logger.debug(f"Applying flatpix2pix: {os.path.basename(infile)} to detector image")
        logger.debug(f"- minimum flatpix2pix value: {np.min(image2d_flatpix2pix):.6f}")
        logger.debug(f"- maximum flatpix2pix value: {np.max(image2d_flatpix2pix):.6f}")
        image2d_detector_method0 /= image2d_flatpix2pix
    else:
        logger.debug("Skipping applying flatpix2pix")

    # Apply bias to detector image
    if bias.value > 0:
        image2d_detector_method0 += bias.value
        logger.debug(f"Applying {bias=} to detector image")
    else:
        logger.debug("Skipping adding bias")

    # Apply Gaussian readout noise to detector image
    if rnoise.value > 0:
        logger.debug(f"Applying Gaussian {rnoise=} to detector image")
        ntot_pixels = naxis1_detector.value * naxis2_detector.value
        image2d_rnoise_flatten = rng.normal(loc=0.0, scale=rnoise.value, size=ntot_pixels)
        image2d_detector_method0 += image2d_rnoise_flatten.reshape((naxis2_detector.value, naxis1_detector.value))
    else:
        logger.debug("Skipping adding Gaussian readout noise")

    save_image2d_detector_method0(
        header_keys,
        image2d_detector_method0=image2d_detector_method0,
        prefix_intermediate_fits=prefix_intermediate_fits,
        bitpix=bitpix_detector,
        logger=logger,
        output_dir=output_dir,
    )

    # ---------------------------------------------------
    # Compute image2d RSS from image in detector, method1
    # ---------------------------------------------------
    image2d_rss_method1 = compute_image2d_rss_from_detector_method1(
        image2d_detector_method0=image2d_detector_method0,
        naxis1_detector=naxis1_detector,
        naxis1_ifu=naxis1_ifu,
        nslices=nslices,
        dict_ifu2detector=dict_ifu2detector,
        wcs3d=wcs3d,
        ncores=ncores,
        logger=logger,
        console=console,
    )

    # Save FITS file
    save_image2d_rss(
        wcs3d=wcs3d,
        header_keys=header_keys,
        image2d_rss=image2d_rss_method1,
        method=1,
        prefix_intermediate_fits=prefix_intermediate_fits,
        bitpix=-32,
        logger=logger,
        output_dir=output_dir,
    )

    # ------------------------------------
    # Compute image3d IFU from RSS method1
    # ------------------------------------
    image3d_ifu_method1 = compute_image3d_ifu_from_rss_method1(
        image2d_rss_method1=image2d_rss_method1,
        naxis1_detector=naxis1_detector,
        naxis2_ifu=naxis2_ifu,
        naxis1_ifu=naxis1_ifu,
        nslices=nslices,
        logger=logger,
    )

    # Save FITS file if prefix is provided
    if len(prefix_intermediate_fits) > 0:
        hdu = fits.PrimaryHDU(image3d_ifu_method1.astype(np.float32))
        pos0 = len(hdu.header) - 1
        hdu.header.extend(wcs_to_header_using_cd_keywords(wcs3d), update=True)
        hdu.header.update(header_keys)
        hdu.header.insert(pos0, ("COMMENT", "FITS (Flexible Image Transport System) format is defined in 'Astronomy"))
        hdu.header.insert(
            pos0 + 1, ("COMMENT", "and Astrophysics', volume 376, page 359; bibcode: 2001A&A...376..359H")
        )
        hdul = fits.HDUList([hdu])
        outfile = f"{prefix_intermediate_fits}_ifu_3D_method1.fits"
        logger.info(f"Saving file: {outfile} (BITPIX=-32)")
        hdul.writeto(f"{Path(output_dir) / outfile}", overwrite=True)



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#
"""Compute the solid angle in arcsec^2 for each pixel in a 2D image."""

import argparse
from argparse import RawTextHelpFormatter
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.wcs import WCS
import numpy as np
from tqdm import tqdm
from rich import print
from scipy.ndimage import median_filter
from spherical_geometry.polygon import SphericalPolygon
import sys

from .add_script_info_to_fits_history import add_script_info_to_fits_history


def pixel_solid_angle_arcsec2(wcs, naxis1, naxis2, method=3, kernel_size=None, verbose=False):
    """Compute the solid angle (arcsec**2) of every pixel.

    This function computes the solid angle for each pixel in a 2D image.
    When using WCS with a very small projected pixel size (e.g., 0.01 arcsec),
    Method 2 is recommended but slow. Method 1 is slow and does not
    work well for very small pixel sizes, but we keep the code here for completeness.
    Method 3 is fast and works well for most cases, but it may not be accurate
    for very small pixel sizes.

    Parameters
    ----------
    wcs : `astropy.wcs.WCS`
        WCS object defining the celestial coordinates of the image.
    naxis1 : int
        Number of pixels along the first axis (NAXIS1).
    naxis2 : int
        Number of pixels along the second axis (NAXIS2).
    method : int, optional
        Method to compute the solid angle.
        1: Use spherical polygons (slow).
        2: Use spherical polygons with a different approach (slow).
        3: Use spherical coordinates and distances (fast, default).
    kernel_size : int, optional
        Size of the kernel for smoothing the result using a median filter.
        If None, no smoothing is applied.
    verbose : bool, optional
        If True, display intermediate information.

    Returns
    -------
    result : `numpy.ndarray`
        2D array with the solid angle (arcsec**2) for each pixel in
        the input image.
    """

    if verbose:
        print(f"Computing solid angle for {naxis1} x {naxis2} pixels using method {method}.")

    # X, Y coordinates (2D image array, following the FITS criterium)
    # corresponding to the four corners of all the image pixels
    borders_x = np.arange(naxis1 + 1) + 0.5
    borders_y = np.arange(naxis2 + 1) + 0.5
    meshgrid = np.meshgrid(borders_x, borders_y)
    ix_array = meshgrid[0].flatten()
    iy_array = meshgrid[1].flatten()

    # spherical coordinates of the four corners of all the image pixels
    result_spherical = SkyCoord.from_pixel(
        xp=ix_array,
        yp=iy_array,
        wcs=wcs,
        origin=1,
        mode='all'
    )

    if method == 1:
        # Use spherical polygons to compute the solid angle
        result_spherical = result_spherical.reshape(naxis2 + 1, naxis1 + 1)
        result = np.zeros((naxis2, naxis1))
        for i in tqdm(range(naxis2), desc="NAXIS2", disable=not verbose):
            for j in range(naxis1):
                polygon = SphericalPolygon.from_radec(
                    lon=[result_spherical[i, j].ra.rad,
                         result_spherical[i, j + 1].ra.rad,
                         result_spherical[i + 1, j + 1].ra.rad,
                         result_spherical[i + 1, j].ra.rad],
                    lat=[result_spherical[i, j].dec.rad,
                         result_spherical[i, j + 1].dec.rad,
                         result_spherical[i + 1, j + 1].dec.rad,
                         result_spherical[i + 1, j].dec.rad],
                    degrees=False,
                )
                result[i, j] = polygon.area() * (180 / np.pi) ** 2 * 3600 ** 2
    elif method == 2:
        # Use spherical polygons with a different approach to compute the solid angle
        result_spherical = result_spherical.reshape(naxis2 + 1, naxis1 + 1)
        result = np.zeros((naxis2, naxis1))
        for i in tqdm(range(naxis2), desc="NAXIS2", disable=not verbose):
            for j in range(naxis1):
                dist1a = result_spherical[i, j].separation(result_spherical[i, j + 1]).arcsec
                dist1b = result_spherical[i + 1, j].separation(result_spherical[i + 1, j + 1]).arcsec
                dist2a = result_spherical[i, j].separation(result_spherical[i + 1, j]).arcsec
                dist2b = result_spherical[i, j + 1].separation(result_spherical[i + 1, j + 1]).arcsec
                # average distances corresponding to opposite sides of each pixel
                dist1 = (dist1a + dist1b) / 2
                dist2 = (dist2a + dist2b) / 2
                # solid angle (arcsec**2) of the pixel
                result[i, j] = dist1 * dist2
    elif method == 3:
        # cartesian coordinates of the four corners of all the image pixels
        x = result_spherical.cartesian.x.value.reshape(naxis2 + 1, naxis1 + 1)
        y = result_spherical.cartesian.y.value.reshape(naxis2 + 1, naxis1 + 1)
        z = result_spherical.cartesian.z.value.reshape(naxis2 + 1, naxis1 + 1)
        # dot product of consecutive points along NAXIS1
        dot_product_naxis1 = x[:, :-1] * x[:, 1:] + \
            y[:, :-1] * y[:, 1:] + z[:, :-1] * z[:, 1:]
        # distance (arcsec) between consecutive points along NAXIS1
        result_naxis1 = np.arccos(dot_product_naxis1) * 180 / np.pi * 3600
        # average distances corresponding to the upper and lower sides of each pixel
        pixel_size_naxis1 = (result_naxis1[:-1, :] + result_naxis1[1:, :]) / 2
        # dot product of consecutive points along NAXIS2
        dot_product_naxis2 = x[:-1, :] * x[1:, :] + \
            y[:-1, :] * y[1:, :] + z[:-1, :] * z[1:, :]
        # distance (arcsec) between consecutive points along NAXIS2
        result_naxis2 = np.arccos(dot_product_naxis2) * 180 / np.pi * 3600
        # average distances corresponding to the left and right sides of each pixel
        pixel_size_naxis2 = (result_naxis2[:, :-1] + result_naxis2[:, 1:]) / 2
        # pixel size (arcsec**2)
        result = pixel_size_naxis1 * pixel_size_naxis2
    else:
        raise ValueError(f"Invalid method: {method}. Choose 1, 2, or 3.")

    # smooth result
    if kernel_size is not None:
        if verbose:
            print(f"Applying median filter with kernel size {kernel_size}.")
        result = median_filter(result, size=kernel_size, mode='nearest')

    return result


def main(args=None):
    parser = argparse.ArgumentParser(
        description="Compute the solid angle in arcsec^2 for each pixel in a 2D image.",
        formatter_class=RawTextHelpFormatter
    )
    parser.add_argument('input_file', type=str,
                        help='FITS file containing the image data.')
    parser.add_argument('output_file', type=str,
                        help='Output FITS file to save the solid angle data.')
    parser.add_argument("--extname", type=str,
                        help="Extension name of the input HDU (default: 'PRIMARY').",
                        default='PRIMARY')
    parser.add_argument("--method", type=int, default=3,
                        help="Method to compute the solid angle:\n"
                             "1: Use spherical polygons (slow)\n"
                             "   (not recommended for very small pixel sizes)\n"
                             "2: Use spherical polygons (different approach, slow)\n"
                             "   (recommended for small pixel sizes)\n"
                             "3: Use spherical coordinates and distances (fast)\n"
                             "   (default, not recommended for very small pixel sizes)")
    parser.add_argument("--kernel_size", type=int, default=None,
                        help="Kernel size for smoothing the result using a median\n"
                             "filter. If not specified, no smoothing is applied.")
    parser.add_argument("--verbose",
                        help="Display intermediate information",
                        action="store_true")
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")

    args = parser.parse_args(args)

    if len(sys.argv) == 1:
        parser.print_usage()
        raise SystemExit()

    if args.verbose:
        for arg, value in vars(args).items():
            print(f'{arg}: {value}')

    if args.echo:
        print('[bold red]Executing:\n' + ' '.join(sys.argv) + '[/bold red]')

    input_file = args.input_file
    output_file = args.output_file
    extname = args.extname
    kernel_size = args.kernel_size
    method = args.method
    verbose = args.verbose

    with fits.open(input_file) as hdul:
        if extname not in hdul:
            raise ValueError(f"Extension '{extname}' not found in {input_file}.")
        hdu_image = hdul[extname]
        naxis = hdu_image.header['NAXIS']
        if naxis == 2:
            wcs = WCS(hdu_image.header)
        elif naxis == 3:
            wcs = WCS(hdu_image.header).celestial
        else:
            raise ValueError(f"Unsupported NAXIS value: {naxis}. Expected 2 or 3.")
        naxis1 = hdu_image.header['NAXIS1']
        naxis2 = hdu_image.header['NAXIS2']
        if args.verbose:
            print(f"{naxis1=}")
            print(f"{naxis2=}")
            print(f"Celestial coordinates WCS:\n{wcs}")

    result = pixel_solid_angle_arcsec2(
        wcs=wcs,
        naxis1=naxis1,
        naxis2=naxis2,
        method=method,
        kernel_size=kernel_size,
        verbose=verbose
    )

    # Create a new FITS HDU with the solid angle data
    solid_angle_hdu = fits.PrimaryHDU(data=result.astype(np.float32))
    header = solid_angle_hdu.header
    add_script_info_to_fits_history(header, args)
    if args.verbose:
        print(f"Saving solid angle data to {output_file}.")
    solid_angle_hdu.writeto(output_file, overwrite=True)



#
# Copyright 2024-2026 Universidad Complutense de Madrid
#
# This file is part of FRIDA DRP
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

from rich.console import Console


def raise_ValueError(msg):
    """Raise exception showing a coloured message."""
    console = Console()
    console.print(msg, style="bold red")
    raise ValueError(msg)



#
# Copyright 2024-2026 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#
import logging

from astropy.io import fits
import numpy as np
from pathlib import Path

from .define_3d_wcs import wcs_to_header_using_cd_keywords


def generate_image3d_method0_ifu(
    wcs3d,
    header_keys,
    simulated_x_ifu_all,
    simulated_y_ifu_all,
    simulated_wave_all,
    bins_x_ifu,
    bins_y_ifu,
    bins_wave,
    prefix_intermediate_fits,
    logger=None,
    output_dir=".",
):
    """Compute 3D image3 IFU, method 0

    The image is calculated by performing a three-dimensional histogram
    over the X, Y coordinates in the IFU and the wavelength of the
    simulated photons.

    The result is saved as a FITS file.

    Parameters
    ----------
    wcs3d : `~astropy.wcs.wcs.WCS`
        WCS of the data cube.
    header_keys : `~astropy.io.fits.header.Header`
        FITS header with additional keywords to be merged together with
        the WCS information.
    simulated_x_ifu_all : `~astropy.units.Quantity`
        Simulated X coordinates of the photons in the IFU.
    simulated_y_ifu_all : `~astropy.units.Quantity`
        Simulated Y coordinates of the photons in the IFU.
    simulated_wave_all : `~astropy.units.Quantity`
        Simulated wavelengths of the photons in the IFU.
    bins_x_ifu : `~numpy.ndarray`
        Bin edges in the naxis1_ifu direction
        (along the slice).
    bins_y_ifu : `~numpy.ndarray`
        Bin edges in the naxis2_ifu direction
        (perpendicular to the slice).
    bins_wave : `~numpy.ndarray`
        Bin edges in the wavelength direction.
    prefix_intermediate_fits : str
        Prefix for output intermediate FITS files. If the length of
        this string is 0, no output is generated.
    logger : `~logging.Logger`, optional
        Logger for logging messages. If `None`, the root logger is used.
    output_dir : str or `~pathlib.Path`, optional
        Output directory to store results. Default is the current directory.
    """
    if logger is None:
        logger = logging.getLogger()

    if simulated_x_ifu_all is None or simulated_y_ifu_all is None or simulated_wave_all is None:
        nphotons_all = 0
    else:
        nphotons_all = len(simulated_x_ifu_all)
        if nphotons_all != len(simulated_y_ifu_all) or nphotons_all != len(simulated_wave_all):
            raise ValueError(
                "The length of simulated_x_ifu_all, simulated_y_ifu_all, and simulated_wave_all must be the same."
            )

    # generate image
    if nphotons_all > 0:
        image3d_method0_ifu, edges = np.histogramdd(
            sample=(simulated_wave_all.value, simulated_y_ifu_all.value, simulated_x_ifu_all.value),
            bins=(bins_wave.value, bins_y_ifu.value, bins_x_ifu.value),
        )
    else:
        image3d_method0_ifu = np.zeros((len(bins_wave) - 1, len(bins_y_ifu) - 1, len(bins_x_ifu) - 1), dtype=int)

    # save FITS file
    if len(prefix_intermediate_fits) > 0:
        outfile = f"{prefix_intermediate_fits}_ifu_3D_method0.fits"
        if image3d_method0_ifu.max() <= 65535:
            bitpix = 16
            hdu = fits.PrimaryHDU(image3d_method0_ifu.astype(np.uint16))
        else:
            # use float to avoid saturation problem
            logger.warning(f"The maximum value in {outfile} is greater than 65535.")
            logger.warning("Saving the image using float32 to avoid saturation.")
            bitpix = -32
            hdu = fits.PrimaryHDU(image3d_method0_ifu.astype(np.float32))
        pos0 = len(hdu.header) - 1
        hdu.header.extend(wcs_to_header_using_cd_keywords(wcs3d), update=True)
        hdu.header.update(header_keys)
        hdu.header.insert(pos0, ("COMMENT", "FITS (Flexible Image Transport System) format is defined in 'Astronomy"))
        hdu.header.insert(
            pos0 + 1, ("COMMENT", "and Astrophysics', volume 376, page 359; bibcode: 2001A&A...376..359H")
        )
        hdul = fits.HDUList([hdu])
        logger.info(f"Saving file: {outfile} (BITPIX={bitpix})")
        hdul.writeto(f"{Path(output_dir) / outfile}", overwrite=True)



#
# Copyright 2008-2026 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""User command line interface of Numina."""


import logging

from .baserun import run_reduce
from .helpers import create_datamanager, load_observations

_logger = logging.getLogger(__name__)


def mode_run_common(args, extra_args, config, mode):
    # FIXME: implement 'recipe' run mode
    if mode == "rec":
        print("Mode not implemented yet")
        return 1
    elif mode == "obs":
        return mode_run_common_obs(args, extra_args, config)
    else:
        raise ValueError(f"Not valid run mode {mode}")


def mode_run_common_obs(args, extra_args, config):
    """Observing mode processing mode of numina."""

    # Loading observation result if exists
    sessions, loaded_obs = load_observations(args.obsresult, args.session)

    # Override like this
    if args.basedir:
        config["tool.run"]["basedir"] = args.basedir
    if args.datadir:
        config["tool.run"]["datadir"] = args.datadir
    if args.calibsdir:
        config["tool.run"]["calibsdir"] = args.calibsdir
    if args.copy_files:
        config["tool.run"]["copy_files"] = str(args.copy_files)
    if args.validate:
        config["tool.run"]["validate"] = str(args.validate)
    if hasattr(args, "strict_reqs"):
        config["tool.run"]["strict_reqs"] = str(args.strict_reqs)

    datamanager = create_datamanager(config, args.reqs, extra_args.extra_control)
    datamanager.backend.add_obs(loaded_obs)

    # Start processing
    jobs = []
    for session in sessions:
        for job in session:
            if job["enabled"] or job["id"] in args.enable:
                jobs.append(job)

    copy_files = config["tool.run"].getboolean("copy_files")
    validate = config["tool.run"].getboolean("validate")
    strict_inputs = config["tool.run"].getboolean("strict_reqs")
    try:
        for job in jobs:
            run_reduce(
                datamanager,
                job["id"],
                copy_files=copy_files,
                validate_inputs=validate,
                validate_results=validate,
                strict_inputs=strict_inputs,
            )
    finally:
        if args.dump_control:
            _logger.debug("dump control status")
            with open("control_dump.yaml", "w") as fp:
                datamanager.backend.dump(fp)



#
# Copyright 2024-2026 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import logging

from astropy.coordinates import SkyCoord
import astropy.units as u
from astropy.wcs.utils import proj_plane_pixel_scales
import matplotlib.pyplot as plt
import numpy as np

from numina.tools.compute_adr_wavelength import compute_adr_wavelength

from .raise_valueerror import raise_ValueError
from .simulate_image2d_from_fitsfile import simulate_image2d_from_fitsfile


def parse_slice_numbers(string):
    """
    Convert interval string to list of integers. This function is
    used to parse the list of slices in the IFU when the geometry type is
    "slices 1,3-5" (for example, to select slices 1, 3, 4, and 5).

    Parameters
    ----------
    string : str
        String containing intervals separated by commas.
        Intervals can be "number1-number2" or single numbers.
        Blank spaces are allowed and ignored.

    Returns
    -------
    list of int
        All integers contained in the intervals.

    Examples
    --------
    >>> parse_slice_numbers("slices 3-7, 12, 14-16")
    [3, 4, 5, 6, 7, 12, 14, 15, 16]
    """
    result = []

    # Remove blank spaces
    string = string.replace(" ", "")

    # Remove the leading "slices" if present
    if string.lower().startswith("slices"):
        string = string[6:]

    # Split by commas and process each interval
    intervals = string.split(",")

    for interval in intervals:
        if "-" in interval:
            # It's a range
            parts = interval.split("-")
            try:
                start = int(parts[0])
                end = int(parts[1])
                result.extend(range(start, end + 1))
            except ValueError:
                raise ValueError(f"Invalid interval: {interval}")
        else:
            # Single number
            try:
                result.append(int(interval))
            except ValueError:
                raise ValueError(f"Invalid number: {interval}")

    return result


def generate_geometry_for_scene_block(
    scene_fname,
    scene_block,
    nphotons,
    apply_seeing,
    seeing_fwhm_arcsec,
    seeing_psf,
    airmass,
    parallactic_angle,
    reference_wave_vacuum_differential_refraction,
    simulated_wave,
    instrument_pa,
    wcs3d,
    min_x_ifu,
    max_x_ifu,
    min_y_ifu,
    max_y_ifu,
    nslices,
    rng,
    logger,
    plots,
):
    """Distribute photons in the IFU focal plane for the scene block.

    Parameters
    ----------
    scene_fname : str
        YAML scene file name.
    scene_block : dict
        Dictonary storing a scene block.
    nphotons : int
        Number of photons to be generated in the scene block.
    apply_seeing : bool
        If True, apply seeing to simulated photons.
    seeing_fwhm_arcsec : `~astropy.units.Quantity`
        Seeing FWHM.
    seeing_psf : str
        Seeing PSF.
    airmass : float
        Airmass.
    parallactic_angle : `~astropy.units.Quantity`
        Parallactic angle. This number must be within the range
        [-90,+90] deg.
    reference_wave_vacuum_differential_refraction : `~astropy.units.Quantity`
        Reference wavelength to compute the differential refraction
        correction. This wavelength corresponds to a correction of
        zero.
    simulated_wave : `~astropy.units.Quantity`
        Array containing `nphotons` simulated photons with the
        spectrum requested in the scene block. These values are
        required to compute the differential refraction correction.
    instrument_pa : `~astropy.units.Quantity`
        Instrument position angle.
    wcs3d : `~astropy.wcs.wcs.WCS`
        WCS of the data cube.
    min_x_ifu : `~astropy.units.Quantity`
        Minimum pixel X coordinate defining the IFU focal plane.
    max_x_ifu : `~astropy.units.Quantity`
        Maximum pixel X coordinate defining the IFU focal plane.
    min_y_ifu : `~astropy.units.Quantity`
        Minimum pixel Y coordinate defining the IFU focal plane.
    max_y_ifu : `~astropy.units.Quantity`
        Maximum pixel Y coordinate defining the IFU focal plane.
    nslices : int
        Number of slices in the IFU.
    rng : `~numpy.random._generator.Generator`
        Random number generator.
    logger : logging.Logger
        Logger for logging messages.
    plots : bool
        If True, plot intermediate results.

    Returns
    -------
    simulated_x_ifu_corrected : `~astropy.units.Quantity`
        Array containing the X coordinate of the 'nphotons' photons
        in the focal plane of the IFU.
    simulated_y_ifu_corrected : `~astropy.units.Quantity`
        Array containing the X coordinate of the 'nphotons' photons
        in the focal plane of the IFU.

    """

    if logger is None:
        logger = logging.getLogger(__name__)

    if len(simulated_wave) != nphotons:
        raise ValueError(f"Unexpected {len(simulated_wave)=} != {nphotons=}")

    factor_fwhm_to_sigma = 1 / (2 * np.sqrt(2 * np.log(2)))

    # plate scale
    plate_scale_x, plate_scale_y, plate_scale_wave = proj_plane_pixel_scales(wcs3d)
    plate_scale_x *= u.deg / u.pix
    plate_scale_y *= u.deg / u.pix

    logger.debug(f"{wcs3d.wcs.cd=}")
    logger.debug(f"{plate_scale_x=}")
    logger.debug(f"{plate_scale_y=}")

    # define geometry type for scene block
    geometry_type = scene_block["geometry"]["type"]

    # simulate photons following the selected geometry
    simulated_x_ifu = None
    simulated_y_ifu = None
    if geometry_type == "flatfield":
        simulated_x_ifu = rng.uniform(low=min_x_ifu.value, high=max_x_ifu.value, size=nphotons)
        simulated_y_ifu = rng.uniform(low=min_y_ifu.value, high=max_y_ifu.value, size=nphotons)
    elif "slices" in geometry_type:
        slice_list = parse_slice_numbers(geometry_type)
        # simulate photons in the selected slices (assuming that the slices are parallel to the X axis and
        # that they cover the whole X range of the IFU)
        slice_width_y = (max_y_ifu - min_y_ifu) / nslices
        nphotons_per_slice = nphotons // len(slice_list)
        remainder_photons = nphotons % len(slice_list)
        if remainder_photons != 0:
            logger.warning(f"Number of photons {nphotons} is not divisible by the number of slices {len(slice_list)}.")
            logger.warning(
                f"Simulating {remainder_photons} extra photons in first slice to reach the number of photons."
            )
        for islice in slice_list:
            lower_y_ifu = min_y_ifu + (islice - 1) * slice_width_y
            upper_y_ifu = min_y_ifu + islice * slice_width_y
            if simulated_x_ifu is None:
                simulated_x_ifu = rng.uniform(
                    low=min_x_ifu.value, high=max_x_ifu.value, size=nphotons_per_slice + remainder_photons
                )
                simulated_y_ifu = rng.uniform(
                    low=lower_y_ifu.value, high=upper_y_ifu.value, size=nphotons_per_slice + remainder_photons
                )
            else:
                simulated_x_ifu = np.concatenate(
                    (simulated_x_ifu, rng.uniform(low=min_x_ifu.value, high=max_x_ifu.value, size=nphotons_per_slice))
                )
                simulated_y_ifu = np.concatenate(
                    (
                        simulated_y_ifu,
                        rng.uniform(low=lower_y_ifu.value, high=upper_y_ifu.value, size=nphotons_per_slice),
                    )
                )
        if len(simulated_x_ifu) != nphotons:
            raise ValueError(f"Unexpected {len(simulated_x_ifu)=} != {nphotons=}")
    elif geometry_type in ["gaussian", "point-like", "from-FITS-image"]:
        if "ra_deg" in scene_block["geometry"]:
            ra_deg = scene_block["geometry"]["ra_deg"]
        else:
            logger.debug("[faint]Assuming ra_deg: 0[/faint]")
            ra_deg = 0.0
        ra_deg *= u.deg
        if "dec_deg" in scene_block["geometry"]:
            dec_deg = scene_block["geometry"]["dec_deg"]
        else:
            logger.debug("[faint]Assuming dec_deg: 0[/faint]")
            dec_deg = 0.0
        dec_deg *= u.deg
        if "delta_ra_arcsec" in scene_block["geometry"]:
            delta_ra_arcsec = scene_block["geometry"]["delta_ra_arcsec"]
        else:
            logger.debug("[faint]Assuming delta_ra_deg: 0[/faint]")
            delta_ra_arcsec = 0.0
        delta_ra_arcsec *= u.arcsec
        if "delta_dec_arcsec" in scene_block["geometry"]:
            delta_dec_arcsec = scene_block["geometry"]["delta_dec_arcsec"]
        else:
            logger.debug("[faint]Assuming delta_dec_deg: 0[/faint]")
            delta_dec_arcsec = 0.0
        delta_dec_arcsec *= u.arcsec
        x_center, y_center = wcs3d.celestial.world_to_pixel(
            SkyCoord(ra=ra_deg + delta_ra_arcsec.to(u.deg), dec=dec_deg + delta_dec_arcsec.to(u.deg))
        )
        # the previous pixel coordinates are assumed to be 0 at the center
        # of the first pixel in each dimension
        x_center += 1
        y_center += 1
        if geometry_type == "point-like":
            simulated_x_ifu = np.repeat(x_center, nphotons)
            simulated_y_ifu = np.repeat(y_center, nphotons)
        elif geometry_type == "gaussian":
            mandatory_keys = ["fwhm_ra_arcsec"]
            for key in mandatory_keys:
                if key not in scene_block["geometry"]:
                    raise_ValueError(f"Expected key '{key}' not found!")
            fwhm_ra_arcsec = scene_block["geometry"]["fwhm_ra_arcsec"] * u.arcsec
            if "fwhm_dec_arcsec" in scene_block["geometry"]:
                fwhm_dec_arcsec = scene_block["geometry"]["fwhm_dec_arcsec"] * u.arcsec
            else:
                fwhm_dec_arcsec = fwhm_ra_arcsec
                logger.debug(f"[faint]Assuming {fwhm_dec_arcsec=}[/faint]")
            if "position_angle_deg" in scene_block["geometry"]:
                position_angle_deg = scene_block["geometry"]["position_angle_deg"] * u.deg
            else:
                position_angle_deg = 0.0 * u.deg
                logger.debug(f"[faint]Assuming {position_angle_deg=}[/faint]")
            # covariance matrix for the multivariate normal
            std_x = fwhm_ra_arcsec * factor_fwhm_to_sigma / plate_scale_x.to(u.arcsec / u.pix)
            std_y = fwhm_dec_arcsec * factor_fwhm_to_sigma / plate_scale_y.to(u.arcsec / u.pix)
            rotation_matrix = np.array(  # note the sign to rotate N -> E -> S -> W
                [
                    [np.cos(position_angle_deg), np.sin(position_angle_deg)],
                    [-np.sin(position_angle_deg), np.cos(position_angle_deg)],
                ]
            )
            covariance = np.diag([std_x.value**2, std_y.value**2])
            rotated_covariance = np.dot(rotation_matrix.T, np.dot(covariance, rotation_matrix))
            # simulate X, Y values
            simulated_xy_ifu = rng.multivariate_normal(mean=[x_center, y_center], cov=rotated_covariance, size=nphotons)
            # compensate for instrument position angle
            simulated_x_ifu = (
                simulated_xy_ifu[:, 0] * np.cos(instrument_pa).value
                + simulated_xy_ifu[:, 1] * np.sin(instrument_pa).value
            )
            simulated_y_ifu = (
                -simulated_xy_ifu[:, 0] * np.sin(instrument_pa).value
                + simulated_xy_ifu[:, 1] * np.cos(instrument_pa).value
            )
        elif geometry_type == "from-FITS-image":
            mandatory_keys = ["filename", "diagonal_fov_arcsec", "background_to_subtract"]
            for key in mandatory_keys:
                if key not in scene_block["geometry"]:
                    raise_ValueError(f"Expected key '{key}' not found!")
            # read reference FITS file
            infile = scene_block["geometry"]["filename"]
            diagonal_fov_arcsec = scene_block["geometry"]["diagonal_fov_arcsec"] * u.arcsec
            background_to_subtract = scene_block["geometry"]["background_to_subtract"]
            # generate simulated locations in the IFU
            simulated_x_ifu_0, simulated_y_ifu_0 = simulate_image2d_from_fitsfile(
                infile=infile,
                diagonal_fov_arcsec=diagonal_fov_arcsec,
                plate_scale_x=plate_scale_x,
                plate_scale_y=plate_scale_y,
                nphotons=nphotons,
                rng=rng,
                background_to_subtract=background_to_subtract,
                plots=plots,
                logger=logger,
            )
            # compensate for instrument rotation angle
            simulated_x_ifu = (
                simulated_x_ifu_0 * np.cos(instrument_pa).value + simulated_y_ifu_0 * np.sin(instrument_pa).value
            )
            simulated_y_ifu = (
                -simulated_x_ifu_0 * np.sin(instrument_pa).value + simulated_y_ifu_0 * np.cos(instrument_pa).value
            )
            # shift image center
            simulated_x_ifu += x_center
            simulated_y_ifu += y_center
        else:
            simulated_x_ifu = None  # avoid PyCharm warning (not aware of raise ValueError)
            simulated_y_ifu = None  # avoid PyCharm warning (not aware of raise ValueError)
            raise_ValueError(f"Unexpected {geometry_type=}")
    else:
        raise_ValueError(f"Unexpected {geometry_type=} in file {scene_fname}")

    # apply seeing
    if apply_seeing:
        if seeing_psf == "gaussian":
            logger.debug(f"Applying Gaussian PSF with {seeing_fwhm_arcsec=}")
            std_x = seeing_fwhm_arcsec * factor_fwhm_to_sigma / plate_scale_x.to(u.arcsec / u.pix)
            simulated_x_ifu += rng.normal(loc=0, scale=abs(std_x.value), size=nphotons)
            std_y = seeing_fwhm_arcsec * factor_fwhm_to_sigma / plate_scale_y.to(u.arcsec / u.pix)
            simulated_y_ifu += rng.normal(loc=0, scale=abs(std_y.value), size=nphotons)
        else:
            raise_ValueError(f"Unexpected {seeing_psf=}")
    else:
        logger.debug("Not applying seeing")

    # apply differential refraction (as a function of wavelength)
    if (geometry_type != "flatfield") and (airmass > 1.0):
        logger.debug("Applying differential refraction correction as a function of wavelength")
        differential_refraction = compute_adr_wavelength(
            airmass=airmass,
            reference_wave_vacuum=reference_wave_vacuum_differential_refraction,
            wave_vacuum=simulated_wave,
            debug=False,
        )
        # compute RA and DEC of each simulated photon
        simulated_coor = wcs3d.celestial.pixel_to_world(simulated_x_ifu - 1.0, simulated_y_ifu - 1.0)
        simulated_ra = simulated_coor.ra
        simulated_dec = simulated_coor.dec
        # apply differential refraction correction (first declination
        # and then right ascension; see Eq. (39) and (40), pp. 71-72,
        # in Textbook on Spherical Astronomy, Smart, 1977).
        simulated_dec += differential_refraction.to(u.deg) * np.cos(parallactic_angle)
        simulated_ra += differential_refraction.to(u.deg) * np.sin(parallactic_angle) / np.cos(simulated_dec)
        # recompute X, Y coordinates in the IFU focal plane
        simulated_x_ifu_corrected, simulated_y_ifu_corrected = wcs3d.celestial.world_to_pixel(
            SkyCoord(ra=simulated_ra, dec=simulated_dec)
        )
        simulated_x_ifu_corrected += 1
        simulated_y_ifu_corrected += 1
        if plots:
            iok = np.argwhere(simulated_wave > 0 * u.m)
            # differential refraction (as a function of wavelength)
            # computed at the center of the IFU
            wmin_simulated = np.min(simulated_wave[iok])
            wmax_simulated = np.max(simulated_wave[iok])
            nsample_wavelengths = 20
            sample_wavelengths = np.linspace(wmin_simulated, wmax_simulated, nsample_wavelengths)
            differential_refraction_center_ifu = compute_adr_wavelength(
                airmass=airmass,
                reference_wave_vacuum=reference_wave_vacuum_differential_refraction,
                wave_vacuum=sample_wavelengths,
            )
            x_center_ifu, y_center_ifu = wcs3d.celestial.wcs.crpix
            simulated_coor = wcs3d.celestial.pixel_to_world(x_center_ifu - 1.0, y_center_ifu - 1.0)
            ra_center_ifu = simulated_coor.ra
            dec_center_ifu = simulated_coor.dec
            ra_center_ifu = np.repeat(ra_center_ifu, nsample_wavelengths)
            dec_center_ifu = np.repeat(dec_center_ifu, nsample_wavelengths)
            dec_center_ifu += differential_refraction_center_ifu.to(u.deg) * np.cos(parallactic_angle)
            ra_center_ifu += (
                differential_refraction_center_ifu.to(u.deg) * np.sin(parallactic_angle) / np.cos(dec_center_ifu)
            )
            x_center_ifu_corrected, y_center_ifu_corrected = wcs3d.celestial.world_to_pixel(
                SkyCoord(ra=ra_center_ifu, dec=dec_center_ifu)
            )
            x_center_ifu_corrected += 1
            y_center_ifu_corrected += 1
            delta_x_center_ifu = x_center_ifu_corrected - x_center_ifu
            delta_y_center_ifu = y_center_ifu_corrected - y_center_ifu
            # differential refraction (as a function of wavelength)
            # computed for each useful simulated photon
            delta_simulated_x_ifu = simulated_x_ifu_corrected - simulated_x_ifu
            delta_simulated_y_ifu = simulated_y_ifu_corrected - simulated_y_ifu
            fig, axarr = plt.subplots(nrows=2, ncols=1, figsize=(6.4, 6.4))
            for iplot in range(2):
                ax = axarr[iplot]
                if iplot == 0:
                    ax.plot(simulated_wave[iok], delta_simulated_x_ifu[iok], ",", label="simulated photons")
                    ax.plot(sample_wavelengths, delta_x_center_ifu, "r+", label="IFU center")
                    ax.set_ylabel(r"$\Delta$simulated_x_ifu (pixel)")
                else:
                    ax.plot(simulated_wave[iok], delta_simulated_y_ifu[iok], ",", label="simulated photons")
                    ax.plot(sample_wavelengths, delta_y_center_ifu, "r+", label="IFU center")
                    ax.set_ylabel(r"$\Delta$simulated_y_ifu (pixel)")
                ax.set_xlabel(f"Wavelength ({simulated_wave.unit})")
                ax.axhline(0, linestyle="--", color="gray")
                ax.axvline(reference_wave_vacuum_differential_refraction.value, linestyle=":", color="C1")
                ax.set_title(f"airmass: {airmass}, parallactic angle: {parallactic_angle}")
                ax.legend()
            plt.tight_layout()
            plt.show()
    else:
        simulated_x_ifu_corrected = simulated_x_ifu
        simulated_y_ifu_corrected = simulated_y_ifu

    # add units
    simulated_x_ifu_corrected *= u.pix
    simulated_y_ifu_corrected *= u.pix

    return simulated_x_ifu_corrected, simulated_y_ifu_corrected



#
# Copyright 2008-2026 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""User command line interface of Numina."""

import os

from .clirundal import mode_run_common


def complete_config(config):
    """Complete config with default values"""
    secname = "tool.run"
    if not config.has_section(secname):
        config.add_section(secname)

    values = {
        "basedir": os.getcwd(),
    }

    for k, v in values.items():
        if not config.has_option(secname, k):
            config.set(secname, k, v)

    return config


def register(subparsers, config):
    complete_config(config)
    parser_run = subparsers.add_parser("run", help="process a observation result")

    parser_run.add_argument(
        "-c",
        "--task-control",
        dest="reqs",
        help="configuration file of the processing task",
        metavar="FILE",
    )
    parser_run.add_argument(
        "-r",
        "--requirements",
        dest="reqs",
        help="alias for --task-control",
        metavar="FILE",
    )
    parser_run.add_argument(
        "-i",
        "--instrument",
        dest="insconf",
        default=None,
        help="name of an instrument configuration",
    )
    parser_run.add_argument(
        "--profile-path",
        dest="profilepath",
        default=None,
        help="location of the instrument profiles",
    )
    parser_run.add_argument(
        "-p",
        "--pipeline",
        dest="pipe_name",
        default="default",
        help="name of a pipeline",
    )
    parser_run.add_argument(
        "--basedir",
        action="store",
        dest="basedir",
        help="path to create the following directories",
    )
    parser_run.add_argument(
        "--datadir",
        action="store",
        dest="datadir",
        help="path to directory containing pristine data",
    )
    parser_run.add_argument(
        "--calibsdir",
        "--rootdir",
        action="store",
        dest="calibsdir",
        help="path to directory containing calibration data",
        metavar="CALIBSDIR",
    )

    parser_run.add_argument(
        "--resultsdir",
        action="store",
        dest="resultsdir",
        help="path to directory to store results",
    )
    parser_run.add_argument(
        "--workdir",
        action="store",
        dest="workdir",
        help="path to directory containing intermediate files",
    )
    parser_run.add_argument(
        "--cleanup",
        action="store_true",
        dest="cleanup",
        default=False,
        help="cleanup workdir on exit [disabled]",
    )
    parser_run.add_argument(
        "--link-files",
        "--not-copy-files",
        action="store_const",
        dest="copy_files",
        const=False,
        help="do not copy observation result and requirement files",
    )
    parser_run.add_argument(
        "-e",
        "--enable",
        action="append",
        default=[],
        metavar="BLOCKID",
        help="enable blocks by id",
    )
    parser_run.add_argument(
        "--dump-control",
        action="store_true",
        help="save the modified task control file",
    )
    parser_run.add_argument(
        "--session", action="store_true", help="use the obresult file as a session file"
    )
    parser_run.add_argument(
        "--validate",
        action="store_const",
        const=True,
        help="validate inputs and results of recipes",
    )

    group_strict = parser_run.add_mutually_exclusive_group()
    group_strict.add_argument(
        "--strict-reqs",
        action="store_true",
        help="end the program if a requirement name is not matched by the recipe",
    )
    group_strict.add_argument(
        "--loose-reqs",
        action="store_false",
        dest="strict_reqs",
        help="warn if a requirement name is not matched by the recipe",
    )
    parser_run.add_argument(
        "obsresult", nargs="+", help="file with the observation result"
    )

    parser_run.set_defaults(command=mode_run_obsmode, strict_reqs=True)

    return parser_run


def mode_run_obsmode(args, extra_args, config):
    mode_run_common(args, extra_args, config, mode="obs")



#
# Copyright 2025-2026 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Execute PyCosmic cosmic ray detection algorithm."""

from importlib.metadata import version

from datetime import datetime
import numpy as np

try:
    import PyCosmic

    PYCOSMIC_AVAILABLE = True
except ModuleNotFoundError as e:
    PYCOSMIC_AVAILABLE = False

from teareduce.cleanest.mergemasks import merge_peak_tail_masks

from .decorated_output import decorate_output


@decorate_output(prompt="PyCosmic > ")
def decorated_pycosmic_det_cosmics(*args, **kwargs):
    """Wrapper for PyCosmic.det_cosmics with decorated output."""
    return PyCosmic.det_cosmics(*args, **kwargs)


@decorate_output(prompt="")
def decorated_merge_peak_tail_masks(*args, **kwargs):
    """Wrapper for merge_peak_tail_masks with decorated output."""
    return merge_peak_tail_masks(*args, **kwargs)


def execute_pycosmic(
    image2d, bool_to_be_cleaned, rlabel_pycosmic, dict_pc_params_run1, dict_pc_params_run2, displaypar, _logger
):
    """Execute PyCosmic cosmic ray detection algorithm."""
    if not PYCOSMIC_AVAILABLE:
        raise ImportError(
            "PyCosmic is not installed. Please install PyCosmic to use\n"
            "the 'pycosmic' or 'mm_pycosmic' crmethod options.\n"
            "You can try installing it via pip:\n"
            "pip install git+https://github.com/nicocardiel/PyCosmic.git@test\n"
        )
    pc_verbose = dict_pc_params_run1["verbose"] or dict_pc_params_run2["verbose"]
    # Determine if 2 runs are needed
    pc_sigma_det_needs_2runs = False
    if dict_pc_params_run1["sigma_det"] != dict_pc_params_run2["sigma_det"]:
        pc_sigma_det_needs_2runs = True
    pc_rlim_needs_2runs = False
    if dict_pc_params_run1["rlim"] != dict_pc_params_run2["rlim"]:
        pc_rlim_needs_2runs = True
    pycosmic_needs_2runs = pc_sigma_det_needs_2runs or pc_rlim_needs_2runs
    if pycosmic_needs_2runs:
        _logger.info("PyCosmic will be run in 2 passes with modified parameters.")
    else:
        _logger.info("PyCosmic will be run in a single pass.")
    # Detect PyCosmic version
    try:
        version_pycosmic = version("PyCosmic")
    except Exception:
        version_pycosmic = "unknown"
    _logger.info(f"using PyCosmic version: {version_pycosmic}")
    # run 1
    if displaypar:
        _logger.info("[green][PyCosmic parameters for run 1][/green]")
        for key in dict_pc_params_run1.keys():
            _logger.info("%s for pycosmic: %s", key, str(dict_pc_params_run1[key]))
    datetime_ini = datetime.now()
    out = decorated_pycosmic_det_cosmics(
        data=image2d, **{key: value for key, value in dict_pc_params_run1.items() if value is not None}
    )
    median2d_pycosmic = out.data
    flag_pc = out.mask.astype(bool)
    # run 2 if needed
    if pycosmic_needs_2runs:
        if displaypar:
            _logger.info("[green][PyCosmic parameters modified for run 2][/green]")
            if pc_sigma_det_needs_2runs:
                _logger.info(
                    "pc_sigma_det for run 2 (run 1): %f (%f)",
                    dict_pc_params_run2["sigma_det"],
                    dict_pc_params_run1["sigma_det"],
                )
            if pc_rlim_needs_2runs:
                _logger.info(
                    "pc_rlim for run 2 (run 1): %f (%f)", dict_pc_params_run2["rlim"], dict_pc_params_run1["rlim"]
                )
        out2 = decorated_pycosmic_det_cosmics(
            data=image2d, **{key: value for key, value in dict_pc_params_run2.items() if value is not None}
        )
        median2d_pycosmic2 = out2.data
        flag_pc2 = out2.mask.astype(bool)
        # combine results from both runs
        flag_pc = decorated_merge_peak_tail_masks(flag_pc, flag_pc2, pc_verbose)
        median2d_pycosmic = median2d_pycosmic2  # use the result from the 2nd run
    flag_pc = np.logical_and(flag_pc, bool_to_be_cleaned)
    flag_pc = flag_pc.flatten()
    datetime_end = datetime.now()
    delta_datetime = datetime_end - datetime_ini
    _logger.info("PyCosmic execution time: %s", str(delta_datetime))
    _logger.info(
        "pixels flagged as cosmic rays by %s: %d (%08.4f%%)",
        rlabel_pycosmic,
        np.sum(flag_pc),
        np.sum(flag_pc) / flag_pc.size * 100,
    )

    return median2d_pycosmic, flag_pc
